# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Handle scdocs manpages
"""

from __future__ import annotations

import shlex
from collections.abc import Sequence
from pathlib import Path

from releaserr.cli_utils import msg
from releaserr.subprocess_util import run

DEFAULT_TO = "markdown_strict+pipe_tables"


def scd2md(
    files: Sequence[Path],
    output_dir: Path | None = None,
    log: bool = False,
    *,
    to: str = DEFAULT_TO,
) -> list[Path]:
    """
    Long, terrible pipeline to convert scd to markdown
    """
    created: list[Path] = []
    for path in files:
        directory = output_dir or path.parent
        base = path.with_suffix("")
        man_type = int(base.suffix.lstrip("."))
        new = directory / f"{base.with_suffix('').name}{man_type}.md"
        if log:
            msg(f"Converting {path} to {new}")
        # fmt: off
        cmd = [
            "sh", "-euo", "pipefail", "-c",
            # Convert scdoc to html
            f"scd2html < {shlex.quote(str(path))}"
            # Remove aria-hidden attributes so pandoc doesn't try to convert them
            "| sed 's|aria-hidden=\"true\"||'"
            "| pandoc --from html "
            # mkdocs doesn't support most of the pandoc markdown extensions.
            # Use markdown_strict and only enable pipe_tables.
            f"--to {shlex.quote(to)}"
            "| sed "
            # Remove anchors that scd2html inserts
            r"-e 's| \[¶\].*||' "
            f"> {shlex.quote(str(new))}",
        ]
        # fmt: on
        run(cmd, log=False)
        created.append(new)
    return created



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Simple, purpose-built release manager for @gotmax23's projects
"""

from __future__ import annotations

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.dev0"

__all__ = ("__version__",)



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Generate project files and create repositories
"""

import asyncio
import readline  # noqa: F401
from pathlib import Path
from typing import List, Optional

import typer
from rich.prompt import Confirm
from rich.rule import Rule
from rich.table import Table
from sourcehut.client import SrhtClient
from typing_extensions import Annotated

from releaserr.cli_utils import CONSOLE, CLIContext, Exit, TyperChoices, err
from releaserr.copr import DEFAULT_CHROOTS, get_copr_client
from releaserr.new.files import (
    BUILD_BACKENDS,
    SPECFILE_TEMPLATES,
    FilesConfig,
    write_files,
)
from releaserr.new.sourcehut import VISIBILITY as Visibility
from releaserr.new.sourcehut import SourcehutProjectMaker


def _guess_license(cli_ctx: CLIContext) -> str:
    project_license = cli_ctx.project_config.get("license")
    # fmt: off
    if (
        isinstance(project_license, dict)
        and isinstance((text := project_license.get("text")), str)
    ):
    # fmt: on
        return text
    elif isinstance(project_license, str):
        return project_license
    return cli_ctx.global_config.default_license


APP = typer.Typer(help=__doc__)


@APP.command()
def files(
    *,
    context: typer.Context,
    path: Annotated[Optional[Path], typer.Argument()] = None,
    name: Annotated[
        Optional[str],
        typer.Option(
            help="Provide name explicitly instead of guessing based on"
            " [PATH]'s basename"
        ),
    ] = None,
    backend: BUILD_BACKENDS = BUILD_BACKENDS.HATCHLING,
    license_id: Annotated[Optional[str], typer.Option("--license")] = None,
    description: Optional[str] = None,
    specfile: Optional[SPECFILE_TEMPLATES] = None,
    matches: Annotated[Optional[List[str]], typer.Option("-m", "--match")] = None,
):
    """
    Generate project files for a Python project
    """
    cli_ctx = context.ensure_object(CLIContext)
    if license_id is None:
        license_id = _guess_license(cli_ctx)
    if path and path != Path("."):
        path.mkdir()
    else:
        description = cli_ctx.project_config.get("description")
        path = Path.cwd()
    if description is None:
        description = input("description: ")
    name = name or path.name
    config = FilesConfig(name, description, path, backend, license_id, specfile)
    write_files(config, matches or ("*",))


class SourcehutResources(TyperChoices):
    GIT = "git"
    LIST = "list"
    TODO = "todo"


@APP.command(name="sourcehut")
def sourcehut_cmd(
    *,
    context: typer.Context,
    description: Optional[str] = None,
    visibility: Visibility = Visibility.PUBLIC,
    dry_run: Annotated[bool, typer.Option("-n", "--dry-run")] = False,
    assume_yes: Annotated[bool, typer.Option("-y", "--assume-yes")] = False,
    resources: Annotated[
        Optional[List[SourcehutResources]],
        typer.Argument(help="[default: all]", show_default=False),
    ] = None,
):
    """
    Create a Sourcehut git repository, mailing list, and tracker
    """

    resources = resources or list(SourcehutResources)
    if dry_run and assume_yes:
        err("Mutually exclusive: --dry-run and --assume-yes")
        raise Exit()

    cli_ctx = context.ensure_object(CLIContext)
    description = (
        description
        if description is not None
        else cli_ctx.project_config.get("description")
    )

    table = Table(show_header=False, title="Input", border_style="red")
    table.add_row("Name", cli_ctx.project_name)
    table.add_row("Description", description)
    table.add_row("Visibility", visibility)
    CONSOLE.print(table, justify="center")

    if (
        not dry_run
        and not assume_yes
        and not Confirm.ask("Create projects?", default=True)
    ):
        err("Aborted.")
        raise Exit(1)

    async def inner() -> None:
        async with SrhtClient.from_config(cli_ctx.global_config.sourcehut) as client:
            project_maker = SourcehutProjectMaker(
                client, cli_ctx.project_name, visibility, description
            )

            CONSOLE.print(Rule(style="blue"))
            protocol = client.protocol
            baseurl = client.baseurl
            table = Table(
                title="Created projects", show_header=False, border_style="green"
            )

            for resource in resources:
                if not dry_run:
                    await getattr(project_maker, f"create_{resource.value}")()
                table.add_row(
                    resource.name.title(),
                    f"{protocol}{resource.value}.{baseurl}/~gotmax23/{cli_ctx.project_name}",
                )
            CONSOLE.print(table, justify="center")
            CONSOLE.print(
                f"To create a new project: {baseurl}/projects/create",
                justify="center",
            )

    asyncio.run(inner())


@APP.command()
def coprs(
    *,
    context: typer.Context,
    chroots: Annotated[
        Optional[List[str]],
        typer.Option(
            "-c", "--chroot", show_default=False, help=f"[default: {DEFAULT_CHROOTS}]"
        ),
    ] = None,
    description: Optional[str] = None,
    specfile: SPECFILE_TEMPLATES = SPECFILE_TEMPLATES.APP,
):
    """
    Create a Copr release and Copr dev Copr for a project
    """

    cli_ctx = context.ensure_object(CLIContext)

    client = get_copr_client()

    baseurl = cli_ctx.global_config.sourcehut.baseurl
    protocol = cli_ctx.global_config.sourcehut.protocol
    for project in (cli_ctx.project_name, cli_ctx.project_name + "-dev"):
        client.project_proxy.add(
            "gotmax23",
            project,
            chroots=chroots or cli_ctx.global_config.copr.default_chroots,
            homepage=f"{protocol}{baseurl}/~gotmax23/{cli_ctx.project_name}",
            contact=f"~gotmax23/{cli_ctx.project_name}@lists.{baseurl}",
            description=description,
        )
    client.package_proxy.add(
        "gotmax23",
        cli_ctx.project_name + "-dev",
        packagename=specfile.package_name(cli_ctx.project_name),
        source_type="scm",
        source_dict={
            "clone_url": f"{protocol}git.{baseurl}/~gotmax23/{cli_ctx.project_name}",
            "scm_type": "git",
            "commitish": "main",
            "source_build_method": "make_srpm",
        },
    )



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import subprocess
from collections.abc import Sequence
from typing import TYPE_CHECKING

from releaserr.cli_utils import ERR_CONSOLE

if TYPE_CHECKING:
    from _typeshed import StrPath


def run(
    cmd: Sequence[StrPath], log: bool = True, **kwargs
) -> subprocess.CompletedProcess:
    kwargs.setdefault("check", True)
    if kwargs.get("capture_output"):
        kwargs.setdefault("text", True)
    if log:
        ERR_CONSOLE.print(f"Running: {tuple(map(str, cmd))}", style="green")
    return subprocess.run(cmd, **kwargs)  # noqa: PLW1510



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Global releaserr config
"""

from __future__ import annotations

import os
from pathlib import Path

from pydantic import BaseModel
from sourcehut.config import SrhtConfig

from releaserr.copr import CoprConfig
from releaserr.toml import load as toml_load

try:
    XDG_CONFIG_HOME = Path(os.environ["XDG_CONFIG_HOME"])
except KeyError:
    XDG_CONFIG_HOME = Path.home() / ".config"

CONFIG_PATH = XDG_CONFIG_HOME / "releaserr.toml"

DEFAULT_LICENSE = "MIT"


class ReleaserrConfig(BaseModel):
    copr: CoprConfig = CoprConfig()
    sourcehut: SrhtConfig = SrhtConfig().read_config()
    default_license: str = DEFAULT_LICENSE

    @classmethod
    def read_config(cls, path: Path | None = None) -> ReleaserrConfig:
        path = path or CONFIG_PATH
        try:
            with path.open("rb") as fp:
                return cls(**toml_load(fp))
        except FileNotFoundError:
            return cls()



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT
from __future__ import annotations

import os
from pathlib import Path

import nox

LINT_FILES = ("releaserr/", "noxfile.py")
LINT_SESSIONS = ("formatters", "codeqa", "typing")
nox.options.sessions = (*LINT_SESSIONS, "integration")


@nox.session
def codeqa(session: nox.Session) -> None:
    session.install(".[codeqa]")
    session.run("ruff", "check", *session.posargs, *LINT_FILES)
    session.run("reuse", "lint")


@nox.session
def formatters(session: nox.Session) -> None:
    session.install(".[formatters]")
    session.run("black", *session.posargs, *LINT_FILES)
    session.run("isort", *session.posargs, *LINT_FILES)


@nox.session
def typing(session: nox.Session) -> None:
    session.install("-e", ".[typing]")
    session.run("mypy", *session.posargs, *LINT_FILES)


@nox.session
def lint(session: nox.Session) -> None:
    for target in LINT_SESSIONS:
        session.notify(target)


def _temp_repo(session: nox.Session) -> None:
    session.run("git", "init", "--initial-branch=main", ".", external=True)
    session.run("git", "add", ".", external=True)
    session.run("git", "commit", "--no-gpg-sign", "-m", "init", external=True)


@nox.session
def integration(session: nox.Session) -> None:
    session.install(".", "nox", "reuse")

    tmp = session.create_tmp()
    project = Path(tmp, "testproj").resolve()
    session.run("rm", "-rf", str(project), external=True)
    with session.chdir(tmp):
        session.run(
            "releaserr", "new", "files", str(project), "--description=test project"
        )
    with session.chdir(project):
        _temp_repo(session)
        session.run("nox", *session.posargs, env={"CI": "1"})


@nox.session
def publish(session: nox.Session) -> None:
    """
    Publish the current git snapshot to PyPI
    """
    session.install(".", "build", "twine")
    session.run("releaserr", "check-remote")
    isolated = [] if "BUILD_SUBMITTER" in os.environ else ["--isolated"]
    session.run("releaserr", "build", *isolated)
    session.run("releaserr", "upload")



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Interact with Fedora Copr
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from pydantic import BaseModel

if TYPE_CHECKING:
    from copr.v3 import Client

DEFAULT_CHROOTS = (
    "epel-9-x86_64",
    "fedora-37-x86_64",
    "fedora-38-x86_64",
    "fedora-39-x86_64",
    "fedora-rawhide-x86_64",
)


class CoprConfig(BaseModel):
    default_chroots: Sequence[str] = DEFAULT_CHROOTS


def get_copr_client() -> Client:
    from copr.v3 import Client

    return Client.create_from_config_file()



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import re
from abc import ABCMeta, abstractmethod
from collections.abc import Sequence
from pathlib import Path
from typing import Any

from releaserr.cli_utils import Exit, err
from releaserr.subprocess_util import run
from releaserr.toml import Reader, Writer
from releaserr.toml import dump as dump_toml
from releaserr.toml import load as load_toml


class VersionHandler(metaclass=ABCMeta):
    changed_files: Sequence[Path] = ()

    def __init__(self, project_name: str, config: dict[str, Any]):
        self.project_name = project_name
        self.config = config
        if hasattr(self, "__post_init__"):
            self.__post_init__()

    @abstractmethod
    def set_version(self, version: str) -> None:
        pass

    def post_version(self) -> str:
        version = self.get_version()
        version += ".post0"
        self.set_version(version)
        return version

    @abstractmethod
    def get_version(self) -> str:
        pass


class HatchVersionHandler(VersionHandler):
    def set_version(self, version: str) -> None:
        run(["hatch", "version", version])

    def get_version(self) -> str:
        return run(["hatch", "version"], capture_output=True).stdout.strip()

    def post_version(self) -> str:
        run(["hatch", "version", "post"], capture_output=True)
        return self.get_version()


class InitPyVersion(VersionHandler):
    PATTERN = re.compile(
        r'(?i)^(__version__|VERSION) *= *([\'"])v?(?P<version>.+?)\2',
        flags=re.MULTILINE,
    )
    version_file: Path

    def __post_init__(self) -> None:
        project_name = self.project_name.replace("-", "_")
        version_file: Path | None = None
        if "version-file" in self.config:
            version_file = Path(self.config["version-file"])
        else:
            for path in (
                Path("src", project_name, "__init__.py"),
                Path(project_name, "__init__.py"),
            ):
                if path.exists():
                    version_file = path
                    break
        if not version_file:
            err("Failed to find version file!")
            raise Exit(1)

        self.version_file: Path = version_file
        self.changed_files = (self.version_file,)

    def get_match(self) -> tuple[re.Match, str]:
        contents = self.version_file.read_text(encoding="utf-8")
        match = self.PATTERN.search(contents)
        if not match:
            err(f"Failed to parse {self.version_file}!")
            raise Exit(1)
        return match, contents

    def get_version(self) -> str:
        match, _ = self.get_match()
        return match["version"]

    def set_version(self, version: str) -> None:
        match, contents = self.get_match()
        start, end = match.span("version")
        self.version_file.write_text(f"{contents[:start]}{version}{contents[end:]}")


class PyprojectTomlVersion(VersionHandler):
    version_file: Path

    def __post_init__(self) -> None:
        self.version_file: Path = Path("pyproject.toml")
        self.changed_files = (self.version_file,)

    def get_version(self) -> str:
        with self.version_file.open("rb") as fp:
            data = load_toml(fp)
        return data["project"]["version"]

    def set_version(self, version: str) -> None:
        with self.version_file.open("rb") as fp:
            data = load_toml(fp, prefered_reader=Reader.TOMLKIT, allow_fallback=False)
        with self.version_file.open("wb") as fp:
            data["project"]["version"] = version
            dump_toml(data, fp, prefered_writer=Writer.TOMLKIT, allow_fallback=False)


VERSION_SOURCES: dict[str, type[VersionHandler]] = {
    "file": InitPyVersion,
    "hatch": HatchVersionHandler,
    "pyproject.toml": PyprojectTomlVersion,
}
